import io
import sys

import numpy as np

from crab_dbg import dbg


def _redirect_stdout_stderr_to_buffer() -> tuple[io.StringIO, io.StringIO]:
    """
    By the nature of dbg(), the only way to test it works is by capture stdout.
    We also capture stderr as dbg() is designed to be compatible with print(), which accepts file=sys.stderr as
    an argument.
    """
    stdout_buffer = io.StringIO()
    stderr_buffer = io.StringIO()

    sys.stdout = stdout_buffer
    sys.stderr = stderr_buffer

    return stdout_buffer, stderr_buffer


def _reset_stdout_stderr():
    sys.stdout = sys.__stdout__
    sys.stderr = sys.__stderr__


def _assert_correct(dbg_outputs: str, expected_outputs: str) -> None:
    """
    This function checks if dbg() outputs desired variable name and their desired value.
    Note that we do not check line no and col no in the output, as they change for almost every modification of this
    file.
    """
    dbg_outputs: list[str] = list(filter(lambda x: len(x) > 0, dbg_outputs.split("\n")))
    expected_outputs: list[str] = list(
        filter(lambda x: len(x) > 0, expected_outputs.split("\n"))
    )

    assert dbg_outputs[0].startswith("[tests/test_dbg.py:")
    assert len(dbg_outputs) == len(expected_outputs)
    for dbg_output, expected_output in zip(dbg_outputs, expected_outputs):
        print(dbg_output, expected_output)
        assert dbg_output.endswith(expected_output)


def test_no_argument():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    dbg()
    _reset_stdout_stderr()

    actual_output = stdout.getvalue().strip()
    assert actual_output.startswith("[tests/test_dbg.py") and actual_output.endswith(
        "]"
    )


def test_single_argument():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    pi = 3.14
    dbg(pi)
    _reset_stdout_stderr()

    _assert_correct(stdout.getvalue(), "pi = 3.14")


def test_print_to_stderr():
    _, stderr = _redirect_stdout_stderr_to_buffer()

    pi = 3.14
    dbg(pi, file=sys.stderr)
    _reset_stdout_stderr()

    _assert_correct(stderr.getvalue(), "pi = 3.14")


def test_single_argument_with_comment():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    pi = 3.14
    dbg(
        pi,  # This comment should not show in dbg output
    )
    _reset_stdout_stderr()

    _assert_correct(stdout.getvalue(), "pi = 3.14")


def test_multiple_arguments():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    pi = 3.14
    ultimate_answer = 42
    flag = True
    stock_price = [100, 99, 101, 1]
    country_to_capital_cities = {
        "China": "Beijing",
        "United Kingdom": "London",
        "Liyue": "Liyue Harbor",
    }
    dbg(
        pi,
        1 + 1,
        sorted(stock_price),
        "This string contains (, ' and ,",
        ultimate_answer,
        flag,  # You can leave a comment here as well, dbg() won't show this comment.
        stock_price,
        country_to_capital_cities,
    )
    _reset_stdout_stderr()

    expected_outputs = """
pi = 3.14
1 + 1 = 2
sorted(stock_price) = [
    1,
    99,
    100,
    101
]
"This string contains (, ' and ," = "This string contains (, ' and ,"
ultimate_answer = 42
flag = True
stock_price = [
    100,
    99,
    101,
    1
]
country_to_capital_cities = {
    China: 'Beijing',
    United Kingdom: 'London',
    Liyue: 'Liyue Harbor'
}
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


def test_cyclic_reference():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    infinite_list = []
    infinite_list.append(infinite_list)

    infinite_dict = {}
    infinite_dict["self"] = infinite_dict
    dbg(infinite_list, infinite_dict)
    _reset_stdout_stderr()

    expected_outputs = """
infinite_list = [
    [...]
]
infinite_dict = {
    self: {...}
}
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


class Phone:
    def __init__(self, brand, color, price):
        self.brand = brand
        self.color = color
        self.price = price

    def __repr__(self):
        return "Phone:\n    Color: %s\n    Brand: %s\n    Price: %s\n" % (
            self.color,
            self.brand,
            self.price,
        )


def test_custom_repr():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    phone = Phone("Apple", "White", 1099)
    dbg(phone)
    _reset_stdout_stderr()

    expected_outputs = """
Phone:
    Color: White
    Brand: Apple
    Price: 1099
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


class Node:
    def __init__(self, val=0, next_=None, prev=None):
        self.val = val
        self.next = next_
        self.prev = prev


class DoubleLinkedList:
    def __init__(self, head=None, tail=None):
        self.head = head
        self.tail = tail

    @staticmethod
    def create(n: int) -> "DoubleLinkedList":
        """
        Create a LinkedList of n elements, value ranges from 0 to n - 1.
        """
        if n <= 0:
            raise ValueError("A double linked list with %d element is meaningless", n)

        head = Node(0)

        cur = head
        for i in range(1, n):
            new_node = Node(i)
            cur.next = new_node
            new_node.prev = cur
            cur = cur.next

        return DoubleLinkedList(head=head, tail=cur)


def test_object():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    double_linke_list = DoubleLinkedList.create(2)
    dbg(double_linke_list)
    _reset_stdout_stderr()

    # Note the CYCLIC REFERENCE here.
    expected_outputs = """
DoubleLinkedList {
    head: Node {
        val: 0
        next: Node {
            val: 1
            next: None
            prev: CYCLIC REFERENCE
        }
        prev: None
    }
    tail: Node {
        val: 1
        next: None
        prev: Node {
            val: 0
            next: CYCLIC REFERENCE
            prev: None
        }
    }
}
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


def test_nested_data_container():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    double_linked_list = DoubleLinkedList.create(2)
    dbg(
        [double_linked_list, double_linked_list],
        (double_linked_list, double_linked_list),
        {"a": 1, "b": [double_linked_list]},
        [
            1,
            2,
            3,
            4,
        ],
    )
    _reset_stdout_stderr()

    expected_outputs = """
[double_linked_list, double_linked_list] = [
    DoubleLinkedList {
        head: Node {
            val: 0
            next: Node {
                val: 1
                next: None
                prev: CYCLIC REFERENCE
            }
            prev: None
        }
        tail: Node {
            val: 1
            next: None
            prev: Node {
                val: 0
                next: CYCLIC REFERENCE
                prev: None
            }
        }
    },
    DoubleLinkedList {
        head: Node {
            val: 0
            next: Node {
                val: 1
                next: None
                prev: CYCLIC REFERENCE
            }
            prev: None
        }
        tail: Node {
            val: 1
            next: None
            prev: Node {
                val: 0
                next: CYCLIC REFERENCE
                prev: None
            }
        }
    }
]
(double_linked_list, double_linked_list) = (
    DoubleLinkedList {
        head: Node {
            val: 0
            next: Node {
                val: 1
                next: None
                prev: CYCLIC REFERENCE
            }
            prev: None
        }
        tail: Node {
            val: 1
            next: None
            prev: Node {
                val: 0
                next: CYCLIC REFERENCE
                prev: None
            }
        }
    },
    DoubleLinkedList {
        head: Node {
            val: 0
            next: Node {
                val: 1
                next: None
                prev: CYCLIC REFERENCE
            }
            prev: None
        }
        tail: Node {
            val: 1
            next: None
            prev: Node {
                val: 0
                next: CYCLIC REFERENCE
                prev: None
            }
        }
    }
)
{'a': 1, 'b': [double_linked_list]} = {
    a: 1,
    b: [
        DoubleLinkedList {
            head: Node {
                val: 0
                next: Node {
                    val: 1
                    next: None
                    prev: CYCLIC REFERENCE
                }
                prev: None
            }
            tail: Node {
                val: 1
                next: None
                prev: Node {
                    val: 0
                    next: CYCLIC REFERENCE
                    prev: None
                }
            }
        }
    ]
}
[1, 2, 3, 4] = [
    1,
    2,
    3,
    4
]
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


class Stack:
    def __init__(self):
        self.data = []

    def push(self, item):
        self.data.append(item)

    def pop(self):
        return self.data.pop()


def test_complex_object():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    double_linked_list = DoubleLinkedList.create(2)
    stack = Stack()
    stack.push(double_linked_list)
    stack.push(double_linked_list)
    dbg(stack)
    _reset_stdout_stderr()

    expected_outputs = """
Stack {
    data: [
        DoubleLinkedList {
            head: Node {
                val: 0
                next: Node {
                    val: 1
                    next: None
                    prev: CYCLIC REFERENCE
                }
                prev: None
            }
            tail: Node {
                val: 1
                next: None
                prev: Node {
                    val: 0
                    next: CYCLIC REFERENCE
                    prev: None
                }
            }
        },
        DoubleLinkedList {
            head: Node {
                val: 0
                next: Node {
                    val: 1
                    next: None
                    prev: CYCLIC REFERENCE
                }
                prev: None
            }
            tail: Node {
                val: 1
                next: None
                prev: Node {
                    val: 0
                    next: CYCLIC REFERENCE
                    prev: None
                }
            }
        }
    ]
}
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


def test_indent_with_multi_line_repr():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    phone = Phone("Apple", "White", 1099)
    dbg({"my_phones": [phone]})
    _reset_stdout_stderr()

    expected_outputs = """
{'my_phones': [phone]} = {
    my_phones: [
        Phone:
            Color: White
            Brand: Apple
            Price: 1099
    ]
}
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


def test_numpy_simple():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    ndarray = np.array([[1, 2, 3], [4, 5, 6]])
    dbg(ndarray)
    _reset_stdout_stderr()

    expected_outputs = """
ndarray = 
array([[1, 2, 3],
       [4, 5, 6]])
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


def test_numpy_nested():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    ndarray = np.array([[1, 2, 3], [4, 5, 6]])
    stack = Stack()
    stack.push({"dict_key": ndarray})
    dbg(stack)
    _reset_stdout_stderr()

    expected_outputs = """
stack = Stack {
    data: [
        {
            dict_key: 
                array([[1, 2, 3],
                       [4, 5, 6]])
        }
    ]
}
"""

    _assert_correct(stdout.getvalue(), expected_outputs)


def test_numpy_as_top_level_associated_variable():
    stdout, _ = _redirect_stdout_stderr_to_buffer()

    ndarray = np.array([[1, 2, 3], [4, 5, 6]])

    class MyClass:
        def __init__(self, np_array):
            self.np_array = np_array

    dbg(MyClass(ndarray))
    _reset_stdout_stderr()

    expected_outputs = """
MyClass(ndarray) = MyClass {
    np_array: 
        array([[1, 2, 3],
               [4, 5, 6]])
}
"""

    _assert_correct(stdout.getvalue(), expected_outputs)



from ast import parse, unparse
import dis
import inspect
from os import path
from sys import stderr
from types import FrameType
from typing import Any


def _is_numpy_tensor_pandas_data(val: Any) -> bool:
    """Check if the value is numpy ndarray, pytorch tensor, or pandas data frame"""

    # Cannot use isinstance() here as import those large library is too time-consuming.
    cls = val.__class__
    module = cls.__module__
    name = cls.__name__

    # Check for numpy.ndarray
    if module == "numpy" and name == "ndarray":
        return True

    # Check for PyTorch Tensor
    if module == "torch" and name == "Tensor":
        return True

    # Check for pandas DataFrame (module starts with 'pandas' and class name is 'DataFrame')
    if module.startswith("pandas.") and name == "DataFrame":
        return True

    return False


def _has_custom_repr(val: Any) -> bool:
    """Check if the value has a custom __repr__."""
    return val.__class__.__repr__ is not object.__repr__


def _has_custom_str(val: Any) -> bool:
    """Check if the value has a custom __str__."""
    return val.__class__.__str__ is not object.__str__


def _is_built_in_types(val: Any) -> bool:
    """
    Determine if a values is an instance of python built-in types.
    """
    return type(val).__module__ == "builtins"


def _get_dbg_raw_args(source_code: str, positions: dis.Positions) -> list[str]:
    """
    Get the arguments to dbg() function as a list of strings. Does not include keyword arguments.
    """
    source_code_lines: list[str] = source_code.split("\n")

    # Get the dbg function call. This might across multiple lines.
    dbg_call_lines: list[str] = list(
        map(
            lambda x: x.strip(),
            source_code_lines[positions.lineno - 1 : positions.end_lineno],
        )
    )

    dbg_call: str = "\n".join(dbg_call_lines)
    dbg_call_args = parse(dbg_call).body[0].value.args
    return [unparse(arg) for arg in dbg_call_args]


def _get_human_readable_repr(obj: Any) -> str:
    """
    Get a useful dbg representation of an object.

    By default, python just prints things like '<__main__.Linkedlist object at 0x102c47560>', which is useless.
    This function returns things like:
    Linkedlist {
        start: Node {
            val: 0,
            next: Node {
                val: 1,
                next: Node {
                    val: 2,
                    next: None,
                }
            }
        }
    }
    """

    def _indent_multiline_str(string: str, indent: int) -> str:
        """
        Apply additional indent to a possibly already formated, multiline representation.
        """
        lines = string.splitlines(keepends=True)
        if not lines:
            return ""

        # No need to indent the first line.
        first_line = lines[0].rstrip("\n\r")
        if len(lines) == 1:
            return first_line

        # Add indent.
        indented_lines = [first_line]
        for line in lines[1:]:
            stripped_line = line.rstrip("\n\r")
            indented_line = " " * indent + stripped_line
            indented_lines.append(indented_line)

        return "\n".join(indented_lines)

    INDENT_INCREMENT = 4

    def _get_human_readable_repr_recursion(
        obj: Any, indent: int, recursion_path: set, ml_container_new_line: bool
    ) -> str:
        """recursion_path: Backtracking algorithm to detect cyclic reference"""
        if id(obj) in recursion_path:
            if isinstance(obj, list):
                return "[...]"
            elif isinstance(obj, dict):
                return "{...}"
            return "CYCLIC REFERENCE"
        recursion_path.add(id(obj))

        try:
            fields_dbg_repr = []

            # Handle data containers.
            if isinstance(obj, (list, set, tuple)):
                for item in obj:
                    # <num_of_ident><val>
                    fields_dbg_repr.append(
                        "%s%s"
                        % (
                            " " * (indent + INDENT_INCREMENT),
                            _get_human_readable_repr_recursion(
                                item,
                                indent + INDENT_INCREMENT,
                                recursion_path,
                                ml_container_new_line=False,
                            ),
                        )
                    )

                if isinstance(obj, list):
                    return (
                        "[\n" + ",\n".join(fields_dbg_repr) + "\n" + " " * indent + "]"
                    )
                elif isinstance(obj, tuple):
                    return (
                        "(\n" + ",\n".join(fields_dbg_repr) + "\n" + " " * indent + ")"
                    )
                else:
                    return (
                        "{\n" + ",\n".join(fields_dbg_repr) + "\n" + " " * indent + "}"
                    )
            elif isinstance(obj, dict):
                for key, value in obj.items():
                    fields_dbg_repr.append(
                        # <num_of_ident><key>: <val>
                        "%s%s: %s"
                        % (
                            " " * (indent + INDENT_INCREMENT),
                            key,
                            _get_human_readable_repr_recursion(
                                value,
                                indent + INDENT_INCREMENT,
                                recursion_path,
                                ml_container_new_line=True,
                            ),
                        )
                    )
                return "{\n" + ",\n".join(fields_dbg_repr) + "\n" + " " * indent + "}"
            elif _is_numpy_tensor_pandas_data(obj):
                if indent == 0:
                    return "\n" + _indent_multiline_str(repr(obj), indent)
                elif ml_container_new_line:
                    return (
                        "\n"
                        + " " * (indent + INDENT_INCREMENT)
                        + _indent_multiline_str(repr(obj), indent + INDENT_INCREMENT)
                    )
                else:
                    return _indent_multiline_str(repr(obj), indent)
            elif _has_custom_repr(obj):
                return _indent_multiline_str(repr(obj), indent)
            elif _has_custom_str(obj) or _is_built_in_types(obj):
                return _indent_multiline_str(str(obj), indent)
            else:
                # Just an object without __repr__ or __str__ provided.
                for key, value in obj.__dict__.items():
                    fields_dbg_repr.append(
                        "%s%s: %s"
                        % (
                            " " * (indent + INDENT_INCREMENT),
                            key,
                            _get_human_readable_repr_recursion(
                                value,
                                indent + INDENT_INCREMENT,
                                recursion_path,
                                ml_container_new_line=True,
                            ),
                        )
                    )
                return (
                    obj.__class__.__name__
                    + " {\n"
                    + "\n".join(fields_dbg_repr)
                    + "\n"
                    + " " * indent
                    + "}"
                )
        finally:
            # This will execute BEFORE any return statement in try block.
            recursion_path.remove(id(obj))

    return _get_human_readable_repr_recursion(obj, 0, set(), ml_container_new_line=True)


def _get_source_code(frame: FrameType, filename: str) -> str | None:
    """
    Get the source code of this frame as a single string.
    We try to read the named file first, if failed, then try inspect.getsource() to handle the cases where source code
    file is not available.
    """
    try:
        with open(filename, "r") as f:
            return f.read()
    except OSError:
        pass

    try:
        return inspect.getsource(frame)
    except OSError:
        return None


def dbg(*evaluated_args, sep=" ", end="\n", file=None, flush=False):
    """
    Python implementation of rust's dbg!() macro. All behaviors should be the same (or similar at least) as dbg!().

    This implementation is meant to be a perfect replacement to python's built-in function print(), so it supports all
    keyword raw_args accepted by print().

    @param sep: Same as print().
    @param end: Same as print().
    @param file: Same as print().
    @param flush: Same as print().
    """
    frame = inspect.currentframe().f_back
    info = inspect.getframeinfo(frame)

    # Cannot use inspect.getsource, limited by dynamic environment, such like pytest.
    source_code = _get_source_code(frame, info.filename)
    if source_code is None:
        print("crab_dbg: Sorry, cannot get original code", file=stderr)
        return

    raw_args = _get_dbg_raw_args(source_code, info.positions)

    assert len(raw_args) == len(evaluated_args), (
        "Number of raw_args does not equal to number of received args"
    )

    # If no arguments at all.
    if len(raw_args) == 0:
        print(
            # [<file_rel_path>:<line_no>:<col_no>]
            "[%s:%s:%s]"
            % (
                path.relpath(info.filename),
                info.lineno,
                info.positions.col_offset + 1,  # Because this is col idx.
            ),
            sep=sep,
            end=end,
            file=file,
            flush=flush,
        )

    for raw_arg, evaluated_arg in zip(raw_args, evaluated_args):
        human_readable_repr = _get_human_readable_repr(evaluated_arg)
        print(
            # [<file_rel_path>:<line_no>:<col_no>] <raw_arg> = <dbg_repr>
            "[%s:%s:%s] %s = %s"
            % (
                path.relpath(info.filename),
                info.lineno,
                info.positions.col_offset + 1,  # Because this is col idx.
                raw_arg,
                human_readable_repr,
            ),
            sep=sep,
            end=end,
            file=file,
            flush=flush,
        )
